"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const topology_compiler = require("../topology_compiler");
const tl = require("../topology_local");
const intf = require("../topology_interfaces");
const log = require("../util/logger");
const serialize_error = require("serialize-error");
/**
 * This class acts as wrapper for local topology when
 * it is run in child process. It handles communication with parent process.
 */
class TopologyLocalWrapper {
    /** Constructor that sets up call routing */
    constructor(proc) {
        let self = this;
        this.process = proc || process;
        this.topology_local = null;
        this.waiting_for_shutdown = false;
        this.log_prefix = "[Wrapper] ";
        this.pingTimeout = 20 * 1000;
        this.pingInterval = 3000;
        this.lastPing = Date.now();
        this.process.on("message", (msg) => {
            self.handle(msg);
        });
        this.process.once("uncaughtException", (e) => {
            log.logger().error(self.log_prefix + "Unhandeled error in topology wrapper");
            log.logger().exception(e);
            self.clearPingInterval();
            self.killProcess(intf.ChildExitCode.unhandeled_error, e);
        });
        this.process.once('SIGINT', () => {
            log.logger().warn(self.log_prefix + "Received SIGINT, this process id = " + self.process.pid);
            if (!self.topology_local) {
                self.exitNonInit("Shutdown", intf.ChildMsgCode.response_shutdown, intf.ChildExitCode.shutdown_notinit_error);
            }
            else {
                self.shutdown();
            }
        });
        this.process.once('SIGTERM', () => {
            log.logger().warn(self.log_prefix + "Received SIGTERM, this process id = " + self.process.pid);
            if (!self.topology_local) {
                self.exitNonInit("Shutdown", intf.ChildMsgCode.response_shutdown, intf.ChildExitCode.shutdown_notinit_error);
            }
            else {
                self.shutdown();
            }
        });
        this.setPingInterval();
    }
    /** sets ping interval */
    setPingInterval() {
        let self = this;
        this.clearPingInterval();
        this.pingIntervalId = setInterval(() => {
            if (!self.process.connected) {
                self.clearPingInterval();
                let s = `${self.log_prefix}Connected property in child process (pid=${self.process.pid}) is false, shutting down topology.`;
                log.logger().error(s);
                // Bad state: we cannot know if there isn't some other parent that's running the same topology.
                // Calling shutdown should be done when we believe the state is OK.
                self.killProcess(intf.ChildExitCode.parent_disconnect, new Error(s));
                return;
            }
            let now = Date.now();
            if (now - this.lastPing > this.pingTimeout) {
                self.clearPingInterval();
                let s = `${self.log_prefix}Ping inside child process (pid=${self.process.pid}) was not received from parent in predefined interval, shutting down topology.`;
                log.logger().error(s);
                // Bad state: we cannot know if there isn't some other parent that's running the same topology.
                // Calling shutdown should be done when we believe the state is OK.
                self.killProcess(intf.ChildExitCode.parent_ping_timeout, new Error(s));
                return;
            }
        }, this.pingInterval);
    }
    /** clears ping interval */
    clearPingInterval() {
        if (this.pingIntervalId) {
            clearInterval(this.pingIntervalId);
            this.pingIntervalId = null;
        }
    }
    /** exit logic when not initialized */
    exitNonInit(fun, msgCode, exitCode) {
        let self = this;
        self.clearPingInterval();
        let s = `${fun} called in the child process, but the topology hasn't been initialized yet.`;
        log.logger().error(self.log_prefix + s);
        self.sendToParent(msgCode, { err: new Error(s) }, () => {
            self.killProcess(exitCode); // error was already sent to parent
        });
    }
    /** Starts infinite loop by reading messages from parent or console */
    start() {
        //let self = this;
    }
    /** Internal main handler for incoming messages */
    handle(msg) {
        let self = this;
        if (msg.cmd === intf.ParentMsgCode.init) {
            if (self.topology_local) {
                let s = `Init called in the child process (${msg.data.general.uuid}), but the topology is already running (${self.topology_local.getUuid()}).`;
                log.logger().error(self.log_prefix + s);
                self.sendToParent(intf.ChildMsgCode.response_init, { err: new Error(s) });
                return;
            }
            self.uuid = msg.data.general.uuid;
            self.log_prefix = `[Wrapper ${self.uuid}] `;
            delete msg.data.general.uuid;
            let compiler = new topology_compiler.TopologyCompiler(msg.data);
            try {
                compiler.compile();
            }
            catch (err) {
                self.sendToParent(intf.ChildMsgCode.response_init, { err: err }, () => {
                    self.killProcess(intf.ChildExitCode.init_error); // error was already sent to parent
                });
                return;
            }
            let topology = compiler.getWholeConfig();
            if (topology.general && topology.general.wrapper) {
                this.pingTimeout = topology.general.wrapper.ping_child_timeout || this.pingTimeout;
                this.pingInterval = topology.general.wrapper.ping_child_interval || this.pingInterval;
                this.setPingInterval();
                if (topology.general.wrapper.log_level) {
                    log.logger().setLevel(topology.general.wrapper.log_level);
                }
            }
            log.logger().important(self.log_prefix + "Initializing topology " + self.uuid);
            // if an internal error is raised we will exit with code 110
            self.topology_local = new tl.TopologyLocal((err) => { self.killProcess(intf.ChildExitCode.internal_error, err); });
            self.topology_local.init(self.uuid, topology, (err) => {
                self.sendToParent(intf.ChildMsgCode.response_init, { err: err }, () => {
                    if (err) {
                        self.killProcess(intf.ChildExitCode.init_error); // error was already sent to parent
                    }
                });
            });
        }
        if (msg.cmd === intf.ParentMsgCode.ping) {
            this.lastPing = Date.now();
            self.sendToParent(intf.ChildMsgCode.response_ping, {});
        }
        if (msg.cmd === intf.ParentMsgCode.run) {
            if (!self.topology_local) {
                let s = `Run called in the child process, but the topology hasn't been initialized yet.`;
                log.logger().error(self.log_prefix + s);
                self.sendToParent(intf.ChildMsgCode.response_run, { err: new Error(s) });
                return;
            }
            self.topology_local.run((err) => {
                self.sendToParent(intf.ChildMsgCode.response_run, { err: err }, () => {
                    if (err) {
                        self.killProcess(intf.ChildExitCode.run_error); // error was already sent to parent
                    }
                });
            });
        }
        if (msg.cmd === intf.ParentMsgCode.pause) {
            if (!self.topology_local) {
                let s = `Pause called in the child process, but the topology hasn't been initialized yet.`;
                log.logger().error(self.log_prefix + s);
                self.sendToParent(intf.ChildMsgCode.response_pause, { err: new Error(s) });
                return;
            }
            self.topology_local.pause((err) => {
                self.sendToParent(intf.ChildMsgCode.response_pause, { err: err }, () => {
                    if (err) {
                        self.killProcess(intf.ChildExitCode.pause_error); // error was already sent to parent
                    }
                });
            });
        }
        if (msg.cmd === intf.ParentMsgCode.shutdown) {
            if (!self.topology_local) {
                self.exitNonInit("Shutdown", intf.ChildMsgCode.response_shutdown, intf.ChildExitCode.shutdown_notinit_error);
                return;
            }
            self.shutdown();
        }
    }
    /** Kill this process the hard way. */
    killProcess(exit_code, err) {
        let self = this;
        self.clearPingInterval();
        if (err) {
            self.sendToParent(intf.ChildMsgCode.error, { err: err }, (e) => {
                if (e) {
                    self.process.exit(intf.ChildExitCode.parent_disconnect);
                }
                else {
                    self.process.exit(exit_code || intf.ChildExitCode.internal_error);
                }
            });
        }
        else {
            // call hard shut-down anyway
            if (this.topology_local) {
                try {
                    this.topology_local.shutdownHard();
                }
                catch (e) {
                    log.logger().error(this.log_prefix + `THIS SHOULD NOT HAPPEN. Error while shutdownHard in topology ${self.uuid}, process id = ${self.process.pid}`);
                    log.logger().exception(e);
                }
            }
            // stop the process after a short while, so that the parent can process the message
            log.logger().important(self.log_prefix + `Calling process.exit(${exit_code || intf.ChildExitCode.exit_ok}) from the child process for topology ${self.uuid}, process id = ${self.process.pid}`);
            self.process.exit(exit_code || intf.ChildExitCode.exit_ok);
        }
    }
    /** This method shuts down the local topology.
     * Any bolt/spout shutdown exception `err` will be propagated
     * to this method and will result in calling self.killProcess(shutdown_internal_error, err)
     */
    shutdown() {
        let self = this;
        try {
            if (self.waiting_for_shutdown) {
                let s = `Ignoring shutdown in the child process (${self.uuid}): the topology is already shutting down (${self.topology_local.getUuid()}).`;
                log.logger().warn(self.log_prefix + s);
                // do not send error to parent (topology not in error state)
                return;
            }
            self.clearPingInterval();
            self.waiting_for_shutdown = true;
            log.logger().important(self.log_prefix + `Shutting down topology ${self.uuid}, process id = ${self.process.pid}`);
            self.topology_local.shutdown((err) => {
                // if we are shutting down due to unrecoverable exception
                // we have the original error from the data field of the message
                self.sendToParent(intf.ChildMsgCode.response_shutdown, { err: err }, () => {
                    if (err) {
                        log.logger().error(self.log_prefix + `Error shutting down topology ${self.uuid}, process id = ${self.process.pid}`);
                        log.logger().exception(err);
                        self.killProcess(intf.ChildExitCode.shutdown_internal_error); // error was already sent to parent
                        return;
                    }
                    self.killProcess(intf.ChildExitCode.exit_ok);
                });
            });
        }
        catch (e) {
            // stop the process if it was not stopped so far
            log.logger().error("THIS SHOULD NOT HAPPEN!"); // topology_local shutdown is never expected to throw (propagate errors through callbacks)
            log.logger().error(this.log_prefix + `Error while shutting down topology ${self.uuid}, process id = ${self.process.pid}`);
            log.logger().exception(e);
            self.sendToParent(intf.ChildMsgCode.response_shutdown, { err: e }, () => {
                self.killProcess(intf.ChildExitCode.shutdown_unlikely_error); // error was already sent to parent
            });
        }
    }
    /** Sends command to parent process.
     * @param {string} cmd - command to send
     * @param {Object} data - data to send
     */
    sendToParent(cmd, data, callback) {
        if (this.process.send) {
            if (data.err) {
                data.err = serialize_error(data.err);
            }
            if (callback) {
                this.process.send({ cmd: cmd, data: data }, null, {}, callback);
            }
            else {
                this.process.send({ cmd: cmd, data: data });
            }
        }
        else {
            // we're running in dev/test mode as a standalone process
            console.log(this.log_prefix + "Sending command", { cmd: cmd, data: data });
            if (callback) {
                return callback();
            }
        }
    }
}
exports.TopologyLocalWrapper = TopologyLocalWrapper;
//# sourceMappingURL=topology_local_wrapper.js.map